/*
 *      Interactive disassembler (IDA).
 *      Copyright (c) 1990-98 by Ilfak Guilfanov.
 *      ALL RIGHTS RESERVED.
 *                              E-mail: ig@estar.msk.su, ig@datarescue.com
 *                              FIDO:   2:5020/209
 *
 */

#include "st6.hpp"

static int ana_basic(void);

//----------------------------------------------------------------------
inline void opA(op_t &op)
{
	op.type = o_reg;
	op.reg = rA;
}

//----------------------------------------------------------------------
inline void opX(op_t &op)
{
	op.type = o_reg;
	op.reg = rX;
}

//----------------------------------------------------------------------
inline void opY(op_t &op)
{
	op.type = o_reg;
	op.reg = rY;
}

//----------------------------------------------------------------------
inline void opV(op_t &op)
{
	op.type = o_reg;
	op.reg = rV;
}

//----------------------------------------------------------------------
inline void opW(op_t &op)
{
	op.type = o_reg;
	op.reg = rW;
}

//----------------------------------------------------------------------
inline void opA(void)
{
	cmd.Op1.type = o_reg;
	cmd.Op1.reg = rA;
}

static uchar opbit(uchar value)
{
	// jrr/jrs use a strange bitmask
	// convert to bit number
	uchar data = (value >> 5);
	switch(data)
	{
		case 1:
			data = 4;
			break;

		case 3:
			data = 6;
			break;

		case 4:
			data = 1;
			break;

		case 6:
			data = 3;
			break;
	}

	return data;
}

//----------------------------------------------------------------------
// analyse a basic instruction
static int ana_basic(void)
{
	sint8 offset;
	uchar opcode;
	uchar code = ua_next_byte();

	uchar nibble0 = (code & 0xF);
	uchar nibble1 = (code >> 4);

	switch( code )
	{
		case 0x35:
			// ld A, X
			cmd.itype = ST6_ld;
			opA();
			opX(cmd.Op2);
			break;

		case 0x75:
			// ld A, Y
			cmd.itype = ST6_ld;
			opA();
			opY(cmd.Op2);
			break;

		case 0xB5:
			// ld A, V
			cmd.itype = ST6_ld;
			opA();
			opV(cmd.Op2);
			break;

		case 0xF5:
			// ld A, W
			cmd.itype = ST6_ld;
			opA();
			opW(cmd.Op2);
			break;

		case 0x3D:
			// ld X, A
			cmd.itype = ST6_ld;
			opX(cmd.Op1);
			opA(cmd.Op2);
			break;

		case 0x7D:
			// ld Y, A
			cmd.itype = ST6_ld;
			opY(cmd.Op1);
			opA(cmd.Op2);
			break;

		case 0xBD:
			// ld V, A
			cmd.itype = ST6_ld;
			opV(cmd.Op1);
			opA(cmd.Op2);
			break;

		case 0xFD:
			// ld W, A
			cmd.itype = ST6_ld;
			opW(cmd.Op1);
			opA(cmd.Op2);
			break;

		case 0x1F:
			// ld A, mem
			cmd.itype = ST6_ld;
			opA();
			cmd.Op2.type = o_mem;
			cmd.Op2.addr = ua_next_byte();
			break;

		case 0x9F:
			// ld mem, A
			cmd.itype = ST6_ld;
			cmd.Op1.type = o_mem;
			cmd.Op1.addr = ua_next_byte();
			opA(cmd.Op2);
			break;

		case 0x07:
			// ld A, (X)
			cmd.itype = ST6_ld;
			opA();
			cmd.Op2.type = o_phrase;
			cmd.Op2.indreg = rX;
			break;

		case 0x0F:
			// ld A, (Y)
			cmd.itype = ST6_ld;
			opA();
			cmd.Op2.type = o_phrase;
			cmd.Op2.indreg = rY;
			break;

		case 0x87:
			// ld (X), A
			cmd.itype = ST6_ld;
			cmd.Op1.type = o_phrase;
			cmd.Op1.indreg = rX;
			opA(cmd.Op2);
			break;

		case 0x8F:
			// ld (Y), A
			cmd.itype = ST6_ld;
			cmd.Op1.type = o_phrase;
			cmd.Op1.indreg = rY;
			opA(cmd.Op2);
			break;

		case 0x0D:
			cmd.itype = ST6_clr;
			opcode = ua_next_byte();
			offset = ua_next_byte();
			
			switch( offset )
			{
				case 0x00:
					switch( opcode )
					{
						case 0x80:
							// clr X
							opX(cmd.Op1);
							break;

						case 0x81:
							// clr Y
							opY(cmd.Op1);
							break;

						case 0x82:
							// clr V
							opV(cmd.Op1);
							break;

						case 0x83:
							// clr W
							opW(cmd.Op1);
							break;

						default:
							// clr mem
							// same as ldi mem, #0
							cmd.Op1.type = o_mem;
							cmd.Op1.addr = opcode;
							break;
					}
					break;

				default:
					// ldi mem, #imm
					cmd.itype = ST6_ldi;
					cmd.Op1.type = o_mem;
					cmd.Op1.addr = opcode;
					cmd.Op2.type = o_imm;
					cmd.Op2.value = offset;
					break;
			}
			break;

		case 0x17:
			// ldi A, #imm
			cmd.itype = ST6_ldi;
			opA();
			cmd.Op2.type = o_imm;
			cmd.Op2.value = ua_next_byte();
			break;

		case 0x04:
			// nop
			cmd.itype = ST6_nop;
			break;

		case 0xED:
			// wait
			cmd.itype = ST6_wait;
			break;

		case 0x6D:
			// stop
			cmd.itype = ST6_stop;
			break;

		case 0x4D:
			// reti
			cmd.itype = ST6_reti;
			break;

		case 0xCD:
			// ret
			cmd.itype = ST6_ret;
			break;

		case 0x15:
			// inc A
			cmd.itype = ST6_inc;
			opX(cmd.Op1);
			break;

		case 0x55:
			// inc Y
			cmd.itype = ST6_inc;
			opY(cmd.Op1);
			break;

		case 0x95:
			// inc V
			cmd.itype = ST6_inc;
			opV(cmd.Op1);
			break;

		case 0xD5:
			// inc W
			cmd.itype = ST6_inc;
			opW(cmd.Op1);
			break;

		case 0x7F:
			cmd.itype = ST6_inc;
			opcode = ua_next_byte();
			switch( opcode )
			{
				case 0xFF:
					// inc A
					opA(cmd.Op1);
					break;

				default:
					// inc mem
					cmd.Op1.type = o_mem;
					cmd.Op1.addr = opcode;
					break;
			}
			break;

		case 0x67:
			// inc (X)
			cmd.itype = ST6_inc;
			cmd.Op1.type = o_phrase;
			cmd.Op1.indreg = rX;
			break;

		case 0x6F:
			// inc (Y)
			cmd.itype = ST6_inc;
			cmd.Op1.type = o_phrase;
			cmd.Op1.indreg = rY;
			break;

		case 0x1D:
			// dec X
			cmd.itype = ST6_dec;
			opX(cmd.Op1);
			break;

		case 0x5D:
			// dec Y
			cmd.itype = ST6_dec;
			opY(cmd.Op1);
			break;

		case 0x9D:
			// dec V
			cmd.itype = ST6_dec;
			opV(cmd.Op1);
			break;

		case 0xDD:
			// dec W
			cmd.itype = ST6_dec;
			opW(cmd.Op1);
			break;

		case 0xFF:
			cmd.itype = ST6_dec;
			opcode = ua_next_byte();
			switch( opcode )
			{
				case 0xFF:
					// dec A
					opA(cmd.Op1);
					break;

				default:
					// dec mem
					cmd.Op1.type = o_mem;
					cmd.Op1.addr = opcode;
					break;
			}
			break;

		case 0xE7:
			// dec (X)
			cmd.itype = ST6_dec;
			cmd.Op1.type = o_phrase;
			cmd.Op1.indreg = rX;
			break;

		case 0xEF:
			// dec (Y)
			cmd.itype = ST6_dec;
			cmd.Op1.type = o_phrase;
			cmd.Op1.indreg = rY;
			break;

		case 0x5F:
			cmd.itype = ST6_add;
			opA(cmd.Op1);
			opcode = ua_next_byte();
			switch( opcode )
			{
				case 0xFF:
					// add A, A
					opA(cmd.Op2);
					break;

				case 0x80:
					// add A, X
					opX(cmd.Op2);
					break;

				case 0x81:
					// add A, Y
					opY(cmd.Op2);
					break;

				case 0x82:
					// add A, V
					opV(cmd.Op2);
					break;

				case 0x83:
					// add A, W
					opW(cmd.Op2);
					break;

				default:
					// add A, mem
					cmd.Op2.type = o_mem;
					cmd.Op2.addr = opcode;
					break;
			}
			break;

		case 0x47:
			// add A, (X)
			cmd.itype = ST6_add;
			opA(cmd.Op1);
			cmd.Op2.type = o_phrase;
			cmd.Op2.indreg = rX;
			break;

		case 0x4F:
			// add A, (Y)
			cmd.itype = ST6_add;
			opA(cmd.Op1);
			cmd.Op2.type = o_phrase;
			cmd.Op2.indreg = rY;
			break;

		case 0x57:
			// addi A, #imm
			cmd.itype = ST6_addi;
			opA(cmd.Op1);
			cmd.Op2.type = o_imm;
			cmd.Op2.value = ua_next_byte();
			break;

		case 0x2D:
			// com A
			cmd.itype = ST6_com;
			opA(cmd.Op1);
			break;

		case 0xAD:
			// rlc A
			cmd.itype = ST6_rlc;
			opA(cmd.Op1);
			break;

		case 0xDF:
			cmd.itype = ST6_sub;
			opA(cmd.Op1);
			opcode = ua_next_byte();
			switch( opcode )
			{
				case 0x80:
					// sub A, X
					opX(cmd.Op2);
					break;

				case 0x81:
					// sub A, Y
					opY(cmd.Op2);
					break;

				case 0x82:
					// sub A, V
					opV(cmd.Op2);
					break;

				case 0x83:
					// sub A, W
					opW(cmd.Op2);
					break;

				case 0xFF:
					// clr A
					// same as sub A, A
					cmd.itype = ST6_clr;
					opA(cmd.Op1);
					break;

				default:
					// sub A, mem
					cmd.Op2.type = o_mem;
					cmd.Op2.addr = opcode;
					break;
			}
			break;

		case 0xBF:
			cmd.itype = ST6_and;
			opA(cmd.Op1);
			opcode = ua_next_byte();
			switch( opcode )
			{
				case 0xFF:
					// and A, A
					opA(cmd.Op2);
					break;

				case 0x80:
					// and A, X
					opX(cmd.Op2);
					break;

				case 0x81:
					// and A, Y
					opY(cmd.Op2);
					break;

				case 0x82:
					// and A, V
					opV(cmd.Op2);
					break;

				case 0x83:
					// and A, W
					opW(cmd.Op2);
					break;

				default:
					// and A, mem
					cmd.Op2.type = o_mem;
					cmd.Op2.addr = opcode;
					break;
			}
			break;

		case 0xA7:
			// and A, (X)
			cmd.itype = ST6_and;
			opA(cmd.Op1);
			cmd.Op2.type = o_phrase;
			cmd.Op2.indreg = rX;
			break;

		case 0xAF:
			// and A, (Y)
			cmd.itype = ST6_and;
			opA(cmd.Op1);
			cmd.Op2.type = o_phrase;
			cmd.Op2.indreg = rY;
			break;

		case 0xB7:
			// andi A, #imm
			cmd.itype = ST6_andi;
			opA(cmd.Op1);
			cmd.Op2.type = o_imm;
			cmd.Op2.value = ua_next_byte();
			break;

		case 0x3F:
			cmd.itype = ST6_cp;
			opA(cmd.Op1);
			opcode = ua_next_byte();
			switch( opcode )
			{
				case 0xFF:
					// cp A, A
					opA(cmd.Op2);
					break;

				case 0x80:
					// cp A, X
					opX(cmd.Op2);
					break;

				case 0x81:
					// cp A, Y
					opY(cmd.Op2);
					break;

				case 0x82:
					// cp A, V
					opV(cmd.Op2);
					break;

				case 0x83:
					// cp A, W
					opW(cmd.Op2);
					break;

				default:
					// cp A, mem
					cmd.Op2.type = o_mem;
					cmd.Op2.addr = opcode;
					break;
			}
			break;

		case 0x27:
			// cp A, (X)
			cmd.itype = ST6_cp;
			opA(cmd.Op1);
			cmd.Op2.type = o_phrase;
			cmd.Op2.indreg = rX;
			break;

		case 0x2F:
			// cp A, (Y)
			cmd.itype = ST6_cp;
			opA(cmd.Op1);
			cmd.Op2.type = o_phrase;
			cmd.Op2.indreg = rY;
			break;

		case 0x37:
			// cpi A, #imm
			cmd.itype = ST6_cpi;
			opA(cmd.Op1);
			cmd.Op2.type = o_imm;
			cmd.Op2.value = ua_next_byte();
			break;

		case 0xC7:
			// sub A, (X)
			cmd.itype = ST6_sub;
			opA(cmd.Op1);
			cmd.Op2.type = o_phrase;
			cmd.Op2.indreg = rX;
			break;

		case 0xCF:
			// sub A, (Y)
			cmd.itype = ST6_sub;
			opA(cmd.Op1);
			cmd.Op2.type = o_phrase;
			cmd.Op2.indreg = rY;
			break;

		case 0xD7:
			// subi A, #imm
			cmd.itype = ST6_subi;
			opA(cmd.Op1);
			cmd.Op2.type = o_imm;
			cmd.Op2.value = ua_next_byte();
			break;

		default:
			opcode = (code & 0x1F);
			switch( opcode )
			{
				case 0x03:
					// jrr #bit, mem, adr
					cmd.itype = ST6_jrr;
					cmd.Op1.type = o_imm;
					cmd.Op1.value = opbit(code);
					cmd.Op2.type = o_mem;
					cmd.Op2.addr = ua_next_byte();
					cmd.Op3.type = o_near;
					offset = (sint8)ua_next_byte();
					cmd.Op3.addr = (cmd.ip+cmd.size) + offset;
					break;

				case 0x13:
					// jrs #bit, mem, adr
					cmd.itype = ST6_jrs;
					cmd.Op1.type = o_imm;
					cmd.Op1.value = opbit(code);
					cmd.Op2.type = o_mem;
					cmd.Op2.addr = ua_next_byte();
					cmd.Op3.type = o_near;
					offset = (sint8)ua_next_byte();
					cmd.Op3.addr = (cmd.ip+cmd.size) + offset;
					break;

				case 0x1B:
					// bset #bit, mem
					cmd.itype = ST6_set;
					cmd.Op1.type = o_imm;
					cmd.Op1.value = opbit(code);
					cmd.Op2.type = o_mem;
					cmd.Op2.addr = ua_next_byte();
					break;

				case 0x0B:
					// bres #bit, mem
					cmd.itype = ST6_res;
					cmd.Op1.type = o_imm;
					cmd.Op1.value = opbit(code);
					cmd.Op2.type = o_mem;
					cmd.Op2.addr = ua_next_byte();
					break;
			}

			switch ( nibble0 )
			{
				case 0x09:
					// jp adr
					cmd.itype = ST6_jp;
					cmd.Op1.type = o_near;
					cmd.Op1.dtyp = dt_word;
					cmd.Op1.addr = ua_next_byte();
					cmd.Op1.addr <<= 4;
					cmd.Op1.addr |= nibble1;
					break;

				case 0x01:
					// call adr
					cmd.itype = ST6_call;
					cmd.Op1.type = o_near;
					cmd.Op1.dtyp = dt_word;
					cmd.Op1.addr = ua_next_byte();
					cmd.Op1.addr <<= 4;
					cmd.Op1.addr |= nibble1;
					break;
			}

			opcode = (nibble0 & 0x07);
			switch( opcode )
			{
				case 0x06:
					// jrc adr
					cmd.itype = ST6_jrc;
					cmd.Op1.type = o_near;
					offset = (sint8)code;
					cmd.Op1.addr = (cmd.ip+cmd.size) + (offset>>3);
					break;

				case 0x02:
					// jrnc adr
					cmd.itype = ST6_jrnc;
					cmd.Op1.type = o_near;
					offset = (sint8)code;
					cmd.Op1.addr = (cmd.ip+cmd.size) + (offset>>3);
					break;

				case 0x04:
					// jrz adr
					cmd.itype = ST6_jrz;
					cmd.Op1.type = o_near;
					offset = (sint8)code;
					cmd.Op1.addr = (cmd.ip+cmd.size) + (offset>>3);
					break;

				case 0x00:
					// jrnz adr
					cmd.itype = ST6_jrnz;
					cmd.Op1.type = o_near;
					offset = (sint8)code;
					cmd.Op1.addr = (cmd.ip+cmd.size) + (offset>>3);
					break;
			}
	}

	return cmd.size;
}

//----------------------------------------------------------------------
// analyse an instruction
int ana(void)
{
	cmd.Op1.dtyp = dt_byte;
	cmd.Op2.dtyp = dt_byte;
	cmd.Op3.dtyp = dt_byte;
	return ana_basic();
}
