/*
 *      Interactive disassembler (IDA).
 *      Copyright (c) 1990-98 by Ilfak Guilfanov.
 *      ALL RIGHTS RESERVED.
 *                              E-mail: ig@estar.msk.su, ig@datarescue.com
 *                              FIDO:   2:5020/209
 *
 */

#include "st6.hpp"

static bool flow;               // does the current instruction pass
                                // execution to the next instruction?

//------------------------------------------------------------------------
// Handle an operand with an immediate value:
//      - mark it with FF_IMMD flag
//      - for bit logical instructions specify the operand type as a number
//        because such an operand is likely a plain number rather than
//        an offset or of another type.

static void doImmdValue(void)
{
	doImmd( cmd.ea );
	switch( cmd.itype )
	{
		case ST6_ldi:
		case ST6_addi:
		case ST6_andi:
		case ST6_cpi:
		case ST6_subi:
		  op_num( cmd.ea, 1 );
		  break;

		case ST6_set:
		case ST6_res:
		case ST6_jrr:
		case ST6_jrs:
			op_num( cmd.ea, 0 );
			break;
	}
	uFlag = getFlags( cmd.ea );		// refresh uFlag with new flags
}

//----------------------------------------------------------------------
// Handle an operand. What this function usually does:
//      - creates cross-references from the operand
//        (the kernel deletes all xrefs before calling emu())
//      - creates permanent comments
//      - if possible, specifies the operand type (for example, it may
//        create stack variables)
//      - anything else you might need to emulate or trace

static void handle_operand( op_t &op,int loading )
{
	switch( op.type )
	{
		case o_phrase:              // no special hanlding for these types
		case o_reg:
			break;

		case o_imm:                         // an immediate number as an operand
			if( !loading )
				goto BAD_LOGIC;				// this can't happen!
			doImmdValue();					// handle immediate number
			break;

		case o_mem:                         // an ordinary memory data reference
		{
			// inf.nametype = NM_SHORT;
			// so we don't need this function anymore
			//char buf[256];
			ea_t dea = intram + op.addr;
			/*segment_t* seg = getseg( dea );
			flags_t flags = getFlags( dea );
			if( seg && !has_user_name( flags))
			{
				sprintf( buf, "%s_%X", get_segm_name( seg ), op.addr );
				set_name( dea, buf, SN_AUTO );
			}*/
			ua_dodata( dea, op.dtyp );
			if( !loading )
				doVar( dea );     // write access
			ua_add_dref( op.offb, dea, loading ? dr_R : dr_W );
		}
		break;

		case o_near:                        // a code reference
		{
			ea_t ea = toEA( cmd.cs, op.addr );
			int iscall = InstrIsSet( cmd.itype, CF_CALL );
			ua_add_cref( op.offb, ea, iscall ? fl_CN : fl_JN );
			if( flow && iscall )
			{
				func_t *pfn = get_func( ea );
				if( pfn != NULL && ( pfn->flags & FUNC_NORET ))
					flow = false;
			}
		}
		break;

		default:
			BAD_LOGIC:
			warning( "%08lX (%s): bad optype", cmd.ea, Instructions[cmd.itype].name );
			break;
	}
}

//----------------------------------------------------------------------
// Emulate an instruction
// This function should:
//      - create all xrefs from the instruction
//      - perform any additional analysis of the instruction/program
//        and convert the instruction operands, create comments, etc.
//      - create stack variables
//      - analyse the delayed branches and similar constructs
// The kernel calls ana() before calling emu(), so you may be sure that
// the 'cmd' structure contains a valid and up-to-date information.
// You are not allowed to modify the 'cmd' structure.
// Upon entering this function, the 'uFlag' variable contains the flags of
// cmd.ea. If you change the characteristics of the current instruction, you
// are required to refresh 'uFlag'.
// Usually the kernel calls emu() with consecutive addresses in cmd.ea but
// you can't rely on this - for example, if the user asks to analyse an
// instruction at arbirary address, his request will be handled immediately,
// thus breaking the normal sequence of emulation.
// If you need to analyse the surroundings of the current instruction, you
// are allowed to save the contents of the 'cmd' structure and call ana().
// For example, this is a very common pattern:
//  {
//    insn_t saved = cmd;
//    if ( decode_prev_insn(cmd.ea) != BADADDR )
//    {
//      ....
//    }
//    cmd = saved;
//  }
//
// This sample emu() function is a very simple emulation engine.

int emu(void)
{
	int Feature = Instructions[cmd.itype].feature;
	flow = ((Feature & CF_STOP) == 0);

	if ( Feature & CF_USE1 )
		handle_operand(cmd.Op1, 1);
	
	if ( Feature & CF_USE2 )
		handle_operand(cmd.Op2, 1);
	
	if ( Feature & CF_USE3 )
		handle_operand(cmd.Op3, 1);
	
	if ( Feature & CF_JUMP )
		QueueMark(Q_jumps,cmd.ea);
	
	if ( Feature & CF_CHG1 )
		handle_operand(cmd.Op1, 0);
	
	if ( Feature & CF_CHG2 )
		handle_operand(cmd.Op2, 0);
	
	if ( Feature & CF_CHG3 )
		handle_operand(cmd.Op3, 0);

	// if the execution flow is not stopped here, then create
	// a xref to the next instruction.
	// Thus we plan to analyse the next instruction.

	if ( flow )
		ua_add_cref( 0, cmd.ea+cmd.size, fl_F );

	return 1;    // actually the return value is unimportant, but let's it be so
}
