/*
 *      Interactive disassembler (IDA).
 *      Copyright (c) 1990-98 by Ilfak Guilfanov.
 *      ALL RIGHTS RESERVED.
 *                              E-mail: ig@estar.msk.su, ig@datarescue.com
 *                              FIDO:   2:5020/209
 *
 */

#include "st6.hpp"
#include <fpro.h>
#include <diskio.hpp>

//----------------------------------------------------------------------
static void vlog(const char *format, va_list va)
{
	static FILE *fp = NULL;
	if ( fp == NULL )
		fp = fopenWT("debug_log");
	qvfprintf(fp, format, va);
	qflush(fp);
}
//----------------------------------------------------------------------
inline void log(const char *format, ...)
{
	va_list va;
	va_start(va, format);
	vlog(format, va);
	va_end(va);
}

//----------------------------------------------------------------------
inline void OutReg(int rgnum)
{
	out_register( ph.regNames[rgnum] );
}

//
// Display an operand that is known to reference data RAM.
//
void outMem(op_t &op)
{
	ea_t        ea;
	const char *ptr;

	//
	// The current data page pointer is known.  Use it to calculate the
	// effective address of the memory being referenced by this
	// operand.
	//
	ea = intram + op.addr;
	 
	//
	// Find or create a name for the data address that this operand
	// references so that we can output that name in the operand's
	// place.
	//

	ptr = get_name_expr( cmd.ea+op.offb, op.n, ea, op.addr );

	//
	// Was a name found or created?
	//
	if( ptr == NULL )
	{
		//
		// No name was found and no name was created.
		// Display the current operand as a regular number.
		//
		OutValue( op, OOF_ADDR );

		//
		// Let the user know that he or she should look at this 
		// instruction and attempt to name the address that it
		// references.
		//
		QueueMark( Q_noName, cmd.ea );
	}
	else
	{
		//
		// A name has been found or created for the address being
		// referenced by this operand.  Display that name in the
		// operand's place.
		//
		OutLine( ptr );
	}
}

// Display an operand that is known to reference another piece of
// of code.
//
void outNear(op_t &op)
{
	ea_t        ea;
	const char *ptr;

	//
	// Calculate the effective address of this code reference.
	//
	ea = toEA( cmd.cs, op.addr );

	//
	// Find or create a name for the code address that this operand
	// references so that we can output that name in the operand's
	// place.
	//
	ptr = get_name_expr( cmd.ea+op.offb, op.n, ea, op.addr );
	if( ptr == NULL )
	{
		//
		// The code address didn't have a name.  Default to
		// displaying the address as a number.
		//
		OutValue( op, OOF_ADDR|OOF_NUMBER|OOFS_NOSIGN );

		//
		// Let the user know that he or she should look at this 
		// instruction and attempt to name the address that it
		// references.
		//
		QueueMark( Q_noName, cmd.ea );
	}
	else
	{
		//
		// The code address either had a name or one has just been
		// created for it.  Display that name in this operand's place.
		//
		OutLine( ptr );
	}
}	

//----------------------------------------------------------------------
// generate the text representation of an operand

int outop(op_t &op)
{
	switch ( op.type )
	{
		case o_reg:
			OutReg( op.reg );
			break;

		case o_phrase:
			out_symbol( '(' );
			OutReg( op.indreg );
			out_symbol( ')' );
			break;

		case o_imm:
			out_symbol( '#' );
			OutValue( op, OOFS_IFSIGN|OOFW_IMM );
			break;

		case o_mem:
			outMem(op);
			break;

		case o_near:
			outNear(op);
			break;

		case o_void:
			return 0;

		default:
			warning( "out: %lx: bad optype",cmd.ea, op.type );
			break;
	}
	return 1;
}

//----------------------------------------------------------------------
// generate a text representation of an instruction
// the information about the instruction is in the 'cmd' structure

void out(void)
{
	char buf[MAXSTR];

	u_line = buf;							// setup the output pointer
	OutMnem();								// output instruction mnemonics

	if( cmd.Op1.type != o_void)
		out_one_operand(0);					// output the first operand

	if( cmd.Op2.type != o_void)
	{
		out_symbol(',');
		OutChar(' ');
		out_one_operand(1);                 // output the second operand
	}

	if( cmd.Op3.type != o_void)
	{
		out_symbol(',');
		OutChar(' ');
		out_one_operand(2);                 // output the third operand
	}


	// output a character representation of the immediate values
	// embedded in the instruction as comments

	if ( isVoid(cmd.ea,uFlag,0) )
		OutImmChar(cmd.Op1);

	if ( isVoid(cmd.ea,uFlag,1) )
		OutImmChar(cmd.Op2);
	
	if ( isVoid(cmd.ea,uFlag,2) )
		OutImmChar(cmd.Op3);

	while ( u_line[-1] == ' ' )
		u_line--;						// trim the trailing spaces

	*u_line = '\0';						// terminate the output string
	gl_comm = 1;						// ask to attach a possible user-
										// defined comment to it
	MakeLine( buf );					// pass the generated line to the
										// kernel
}

//--------------------------------------------------------------------------
// generate start of the disassembly

void header(void)
{
	gen_cmt_line( "Processor:        %s", inf.procName );
	gen_cmt_line( "Device:           %s", device );
	gen_cmt_line( "Target assembler: %s", ash.name );
	if( ash.header != NULL )
	{
		for ( const char **ptr=ash.header; *ptr != NULL; ptr++ )
			MakeLine( *ptr,0 );
	}
}

//--------------------------------------------------------------------------
// generate start of a segment

void segstart(ea_t ea)
{
	char buf[MAXSTR];
	char buf2[MAXSTR];
	// generate ORG directive if necessary
	if (inf.s_org) {
		// get segment data
		segment_t *Sarea = getseg(ea);
		ulong org = get_segm_base(Sarea);
		if( Sarea->type == SEG_CODE )
		{
			// generate line
			sprintf(buf2, "%s %s", ash.origin, btoa(org));
			tag_addstr(buf, COLOR_ASMDIR, buf2);
			MakeLine(buf);
		}
	}
}

//--------------------------------------------------------------------------
// generate end of the disassembly

void footer(void)
{
	char buf[MAXSTR];
	if ( ash.end != NULL)
	{
		MakeNull();
		tag_addstr(buf, COLOR_ASMDIR, ash.end);
		MakeLine(buf);
	}
	else
	{
		gen_cmt_line( "end of file" );
	}
}

//--------------------------------------------------------------------------
// generate a data representation
// usually all the job is handled by the kernel's standard procedure,
// intel_data()

void ST6_data(ea_t ea)
{
	gl_name = 1;
	gl_bpsize = 0;

	segment_t *seg = getseg(ea);

	// output data for our internal ram
	if ( seg != NULL && seg->type == RAMSEG )
	{
		char buf[MAXSTR];
		register char *p;

		gl_name = 0;

		if( get_colored_name(BADADDR, ea, buf, sizeof(buf) ) != NULL )
		{
			uchar off = uchar(ea - get_segm_base(seg));
			p = &buf[strlen(buf)];
			*p++ = ' ';
			p = tag_addstr( p, COLOR_KEYWORD, ash.a_equ );
			*p++ = ' ';
			tag_addstr(p, COLOR_NUMBER, btoa(off));
			MakeLine(buf, 0);
		}
		else
		{
			MakeLine("");
		}
	}
	else
	{
		// use standard ida function to output data
		intel_data( ea );
	}
}
