/*
 *      Interactive disassembler (IDA).
 *      Copyright (c) 1990-98 by Ilfak Guilfanov.
 *      ALL RIGHTS RESERVED.
 *                              E-mail: ig@estar.msk.su, ig@datarescue.com
 *                              FIDO:   2:50620/209
 *
 */

#include "st6.hpp"
#include <entry.hpp>
#include <srarea.hpp>

//--------------------------------------------------------------------------

ea_t intram = 0;

#ifdef _DEBUG
static const insn_t* pinstruction = &cmd;
static const idainfo* pprogram = &inf;
static const processor_t* pprocessor = &ph;
static const asm_t* passembler = &ash;
static const areacb_t* pSeg = &segs;
#endif

static char *RegNames[] =
{
	"A", "V", "W","X","Y","C","Z","PC",
	"cs","ds"
};

//----------------------------------------------------------------------
static ioport_t *ports = NULL;
static int numports = 0;
char device[MAXSTR] = "";
static const char cfgname[] = "st6.cfg";

static clean_symbols(void)
{
	ea_t oldea;

	// delete old symbols first
	// between 0xC0 and 0xFE
	for( int i=0xC0; i < 0xFF; i++ )
	{
		oldea = intram + i;
		flags_t flags = getFlags( oldea );
		if( has_any_name( flags ))
			del_global_name( oldea );
	}

	// delete interrupt vectors
	for( int adr=0xFA0; adr <= 0xFFF; adr++ )
	{
		flags_t flags = getFlags( adr );
		if( has_any_name( flags ))
			del_global_name( adr );
	}
}

static void load_symbols(char* pDevice)
{
	// read our config file - st6.cfg
	free_ioports(ports, numports);
	ports = read_ioports(&numports, cfgname, pDevice, NULL);
}

//----------------------------------------------------------------------
static void apply_symbols(int delSymbols = false)
{
	if( delSymbols )
		clean_symbols();

	for ( int i=0; i < numports; i++ )
	{
		ioport_t &p = ports[i];
		ea_t ea = p.address;
		if( ea < 0x100 )
			ea += intram;
		ea_t oldea = get_name_ea( BADADDR, p.name );
		if( oldea != ea )
		{
			if( oldea != BADADDR )
				del_global_name( oldea );
			do_unknown(ea, 1);
			set_name(ea, p.name, SN_PUBLIC);
		}
		if( p.cmt != NULL )
			set_cmt( ea,p.cmt, 1 );
	}

	// add our processor registers
	// theses are actually ram based (0x80-0x83)
	set_name(intram+0xFF, "A", SN_PUBLIC|SN_NOLIST);
	set_name(intram+0x80, "X", SN_PUBLIC|SN_NOLIST);
	set_name(intram+0x81, "Y", SN_PUBLIC|SN_NOLIST);
	set_name(intram+0x82, "V", SN_PUBLIC|SN_NOLIST);
	set_name(intram+0x83, "W", SN_PUBLIC|SN_NOLIST);
}

//----------------------------------------------------------------------
// Get linear address of a special segment
//      sel - selector of the segment
static long specialSeg(sel_t sel)
{
  segment_t *s = get_segm_by_sel(sel);
  if ( s != NULL )
  {
    if ( s->type != RAMSEG )          // is the segment type correct? - no
    {
      s->type = RAMSEG;               // fix it
      s->update();
    }
    return s->startEA;
  }
  return BADADDR;
}

//----------------------------------------------------------------------
static ea_t AdditionalSegment(int size,int offset,char *name)
{
	segment_t s;
	s.startEA	= (inf.maxEA + 0xF) & ~0xF;
	s.endEA		= s.startEA + size;
	s.sel		= allocate_selector((s.startEA-offset) >> 4);
	s.type		= RAMSEG;
	add_segm_ex(&s, name, NULL, ADDSEG_NOSREG|ADDSEG_OR_DIE);
	return s.startEA - offset;
}

//----------------------------------------------------------------------
// The kernel event notifications
// Here you may take desired actions upon some kernel events

static int notify(processor_t::idp_notify msgid, ...)
{
	va_list va;
	va_start( va, msgid );

	// A well behaving processor module should call invoke_callbacks()
	// in his notify() function. If this function returns 0, then
	// the processor module should process the notification itself
	// Otherwise the code should be returned to the caller:

	int code = invoke_callbacks( HT_IDP, msgid, va );
	if ( code )
		return code;

	switch(msgid)
	{
		case ph.init:
			inf.mf = 1;       // Set a big endian mode of the IDA kernel
			break;

		case ph.newfile:
		{
			// change to short names
			inf.nametype = NM_SHORT;

			// create segments for ram
			segment_t *scode = getnseg(0);
			set_segm_class(scode, "CODE");
			set_segm_name( scode, "CODE" );

			intram = AdditionalSegment( 256, 0, "RAM" );
			
			set_default_dataseg(getseg(intram)->sel);

			// allow user to change processor
			// from list
			if( choose_ioport_device( cfgname, device, NULL ))
				load_symbols(device);
			// apply default symbols
			apply_symbols();
		}
		break;

		case ph.oldfile:
			sel_t sel;
			if ( atos("RAM",&sel) )
				intram = specialSeg(sel);
			break;

		case ph.newseg:
			// make the default DS point to RAM
			{
			segment_t *newseg = va_arg(va, segment_t *);
			segment_t *intseg = getseg(intram);
			if ( intseg != NULL )
				newseg->defsr[rVds-ph.regFirstSreg] = intseg->sel;
			}
			break;

		case ph.newprc:
			load_symbols(device);
			break;

		default:
			break;
	}

	va_end( va );

	return (1);
}

static const char* set_options(const char *keyword,int value_type,const void *value)
{
	char old_device[MAXSTR];
	strcpy( old_device, device );

	// we use this callback so user can change selected micro#
	if( choose_ioport_device( cfgname, device, NULL ))
	{
		if( strcmp( old_device, device ) != 0 )
		{
			load_symbols(device);
			// apply default symbols
			int delSymbols = askyn_c( 1, "%s", "Would you like to Delete the Current Symbols ?" );
			apply_symbols(( delSymbols  == 1 ));
		}
	}
	return IDPOPT_OK;
}

//-----------------------------------------------------------------------
//                   ASMI
//-----------------------------------------------------------------------
static asm_t ast6 = {
	AS_NOCODECLN|ASH_HEXF0|ASB_BINF0|AS_1TEXT|AS_NCHRE|ASO_OCTF0,
	UAS_DEF,
	"SGS Thomson AST6",	// name
	0,					// help
	NULL,				// no headers
	NULL,				// no bad instructions
	".org",				// org directive
	".end",				// end directive
	";",				// comment string
	'"',				// string delimiter
	'\'',				// char delimiter
	"\\\"'",			// special symbols in char and string constants
	".ascii",			// ascii string directive
	".byte",			// byte directive
	".word",			// word directive
	NULL,				// dword  (4 bytes)
	NULL,				// qword  (8 bytes)
	NULL,				// oword  (16 bytes)
	NULL,				// float  (4 bytes)
	NULL,				// double (8 bytes)
	NULL,				// tbyte  (10/12 bytes)
	NULL,				// packed decimal real
	NULL,				// arrays (#h,#d,#v,#s(...)
	".block %s",		// uninitialized arrays
	".equ",				// equ
	NULL,				// seg prefix
	NULL, NULL, NULL,
	NULL,
	"$",				// current ip
	NULL,				// func_header
	NULL,				// func_footer
	NULL,				// public
	NULL,				// weak
	NULL,				// extrn
	NULL,				// comm
	NULL,				// get_type_name
	NULL,				// align
	'(', ')',			// lbrace, rbrace
	"%",				// mod
	"&",				// and
	"|",				// or
	"^",				// xor
	"~",				// not
	"<<",				// shl
	">>",				// shr
	NULL,				// sizeof
	NULL,				// flag2
	NULL,				// cmnt2
	NULL,				// low8
	NULL,				// high8
	NULL,				// low16
	NULL				// high16
};

static asm_t rkitst6 = {
	AS_COLON|ASH_HEXF0|ASB_BINF0|AS_1TEXT|AS_NCHRE|ASO_OCTF0|AS_RELSUP,
	NULL,
	"Raisonance MAST6",	// name
	0,					// help
	NULL,				// no headers
	NULL,				// no bad instructions
	"org",				// org directive
	"end",				// end directive
	";",				// comment string
	'"',				// string delimiter
	'\'',				// char delimiter
	"\\\"'",			// special symbols in char and string constants
	"dcb",				// ascii string directive
	"dsb",				// byte directive
	"dsw",				// word directive
	NULL,				// dword  (4 bytes)
	NULL,				// qword  (8 bytes)
	NULL,				// oword  (16 bytes)
	NULL,				// float  (4 bytes)
	NULL,				// double (8 bytes)
	NULL,				// tbyte  (10/12 bytes)
	NULL,				// packed decimal real
	NULL,				// arrays (#h,#d,#v,#s(...)
	"dcb %s",			// uninitialized arrays
	"equ",				// equ
	NULL,				// seg prefix
	NULL, NULL, NULL,
	NULL,
	"$",				// current ip
	NULL,				// func_header
	NULL,				// func_footer
	NULL,				// public
	NULL,				// weak
	NULL,				// extrn
	NULL,				// comm
	NULL,				// get_type_name
	NULL,				// align
	'(', ')',			// lbrace, rbrace
	"MOD",				// mod
	"AND",				// and
	"OR",				// or
	"XOR",				// xor
	"NOT",				// not
	"SHL",				// shl
	"SHR",				// shr
	NULL,				// sizeof
	NULL,				// flag2
	NULL,				// cmnt2
	NULL,				// low8
	NULL,				// high8
	NULL,				// low16
	NULL				// high16
};

static asm_t *asms[] = { &rkitst6, &ast6, NULL };

//-----------------------------------------------------------------------
// The short and long names of the supported processors
// The short names must match
// the names in the module DESCRIPTION in the makefile (the
// description is copied in the offset 0x80 in the result DLL)

static char *shnames[] =
{
	"st6",
	NULL
};

static char *lnames[] =
{
	"SGS Thomson ST6",
	NULL
};

//--------------------------------------------------------------------------
// Opcodes of "return" instructions. This information will be used in 2 ways:
//      - if an instruction has the "return" opcode, its autogenerated label
//        will be "locret" rather than "loc".
//      - IDA will use the first "return" opcode to create empty subroutines.

static uchar retcode_1[] = { 0xCD };	// ret
static uchar retcode_2[] = { 0x4D };	// reti

static bytes_t retcodes[] = {
	{ sizeof(retcode_1), retcode_1 },
	{ sizeof(retcode_2), retcode_2 },
	{ 0, NULL }                            // NULL terminated array
};

//-----------------------------------------------------------------------
//      Processor Definition
//-----------------------------------------------------------------------
extern "C" processor_t LPH = {
	IDP_INTERFACE_VERSION,// version
	0x8001,	            // id
	PR_RNAMESOK         // can use register names for byte names
	|PR_BINMEM,         // The module creates RAM/ROM segments for binary files
						// (the kernel shouldn't ask the user about their sizes and addresses)
	8,                  // 8 bits in a byte for code segments
	8,                  // 8 bits in a byte for other segments

	shnames,            // array of short processor names
						// the short names are used to specify the processor
						// with the -p command line switch)
	lnames,             // array of long processor names
						// the long names are used to build the processor type
						// selection menu

	asms,               // array of target assemblers

	notify,             // the kernel event notification callback

	header,             // generate the disassembly header
	footer,             // generate the disassembly footer

	segstart,           // generate a segment declaration (start of segment)
	std_gen_segm_footer,// generate a segment footer (end of segment)

	NULL,               // generate 'assume' directives

	ana,                // analyse an instruction and fill the 'cmd' structure
	emu,                // emulate an instruction

	out,                // generate a text representation of an instruction
	outop,              // generate a text representation of an operand
	ST6_data,			// generate a text representation of a data item
	NULL,				// compare operands
	NULL,				// can an operand have a type?

	qnumber(RegNames),	// Number of registers
	RegNames,			// Regsiter names
	NULL,				// get abstract register

	0,					// Number of register files
	NULL,				// Register file names
	NULL,				// Register descriptions
	NULL,				// Pointer to CPU registers

	rVcs,rVds,
	0,					// size of a segment register
	rVcs,rVds,

	NULL,				// No known code start sequences
	retcodes,

	0,ST6_last,
	Instructions,
	NULL,
	NULL,
	NULL,
	NULL,
	{NULL,NULL,NULL,NULL},
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	set_options,
	NULL,
	NULL
};
